import 'dart:io';
import 'dart:mirrors';
import 'package:args/args.dart';
import 'package:container/container.dart';
import 'package:json_god/json_god.dart' as god;
import 'src/metadata.dart';
import 'src/task_info.dart';
import 'io.dart';

final Container _container = new Container();

runRapier([List<String> args]) {
  final List<String> _args = args ?? [];

  final parser = new ArgParser(allowTrailingOptions: true)
    ..addFlag('debug',
        help: 'Print verbose debug output.',
        defaultsTo: false,
        negatable: false)
    ..addFlag('help',
        abbr: 'h',
        help: 'Print this help information.',
        defaultsTo: false,
        negatable: false);

  final result = parser.parse(_args);

  if (result['help']) {
    return exit(0);
  }

  if (result.rest.isEmpty) {
    return runAllTasks(debug: result['debug']);
  } else
    return runTasks(result.rest, debug: result['debug']);
}

runAllTasks({bool debug: false}) => _runTasks(_findTasks(), debug: debug);

runTasks(List<String> names, {bool debug: false}) async {
  final tasks = _findTasks();

  for (String name in names) {
    final task =
        tasks.firstWhere((task) => task.name == name, orElse: () => null);

    if (task == null) {
      stderr.writeln('Could not find a task named "$name".');
      return exit(1);
    } else
      tasks.add(task);
  }

  return _runTasks(tasks, debug: debug);
}

Iterable<Task> _discoverTaskMetadata(List<InstanceMirror> metadata) => metadata
    .map((meta) => meta.reflectee)
    .where((meta) => meta is Task)
    .where((Task task) => !task.skip);

List<TaskInfo> _findTasks() {
  final List<TaskInfo> tasks = [];

  currentMirrorSystem().libraries.forEach((Uri uri, LibraryMirror lib) {
    _registerTasks(lib, 'lib:$uri', lib.declarations, lib.metadata, tasks);
  });

  return tasks;
}

_registerTasks(
    declaration,
    String name,
    Map<Symbol, DeclarationMirror> declarations,
    List<InstanceMirror> metadata,
    List<TaskInfo> tasks,
    {bool debug: false,
    bool skipMeta: false}) {
  final discovered = _discoverTaskMetadata(metadata)
      .map((Task task) => new TaskInfo(task, 'All tasks in $name', (_) {
            final tasks = [];
            _registerTasks(declaration, name, declarations, metadata, tasks,
                debug: debug, skipMeta: true);
            return _runTasks(tasks, debug: debug || task.debug);
          }));

  if (!skipMeta && discovered.isNotEmpty) {
    tasks.addAll(discovered);
  } else {
    declarations.forEach((Symbol className, DeclarationMirror decl) {
      if (decl is MethodMirror) {
        final discovered = _discoverTaskMetadata(decl.metadata)
            .map((Task task) => new TaskInfo(task, task.name, (rapier) {
                  if (decl.isStatic) {
                    return declaration
                        .invoke(
                            className, [rapier]..addAll(task.args), task.named)
                        .reflectee;
                  } else {
                    final instance = _container.make(declaration.reflectedType);
                    final instanceMirror = reflect(instance);

                    final named = {};

                    task.named.forEach((k, v) {
                      named[new Symbol(k)] = v;
                    });

                    return instanceMirror
                        .invoke(className, task.args, named)
                        .reflectee;
                  }
                }));

        tasks.addAll(discovered);
      } else if (decl is ClassMirror) {
        final name = MirrorSystem.getName(decl.simpleName);
        _registerTasks(decl, name, decl.declarations, decl.metadata, tasks,
            debug: debug);
      }
    });
  }
}

_runTasks(Iterable<TaskInfo> tasks, {bool debug: false}) async {
  for (TaskInfo task in tasks) {
    final willDebug = debug || task.source.debug;

    if (willDebug) print('Now running "${task.name}"...');

    final rapier = new Rapier(debug: willDebug);
    _container.singleton(rapier, as: BaseRapier);
    _container.singleton(rapier, as: Rapier);

    try {
      final result = await task.run(rapier);

      if (result != null) {
        var toPrint = result;

        switch (task.source.out) {
          case Task.JSON:
            toPrint = god.serialize(result);
            break;
        }

        if (willDebug) {
          print('Result: $toPrint');
        } else
          print(toPrint);
      }
    } catch (e, st) {
      if (willDebug) {
        stderr.writeln('Task "${task.name}" failed.');
      }

      stderr.writeln(e);
      stderr.writeln(st);
      return exit(1);
    }
  }

  return exit(0);
}
